# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestMerging(unittest.TestCase):
    def test_simple_merging_solver(self):
        self.raw_simple_merging(claripy.Solver)

    def test_simple_merging_solverhybrid(self):
        self.raw_simple_merging(claripy.SolverHybrid)

    def test_simple_merging_solvercomposite(self):
        self.raw_simple_merging(claripy.SolverComposite)

    def raw_simple_merging(self, solver_type):
        s1 = solver_type()
        s2 = solver_type()
        w = claripy.BVS("w", 8)
        x = claripy.BVS("x", 8)
        y = claripy.BVS("y", 8)
        z = claripy.BVS("z", 8)
        m = claripy.BVS("m", 8)

        s1.add([x == 1, y == 10])
        s2.add([x == 2, z == 20, w == 5])
        _, sm = s1.merge([s2], [m == 0, m == 1])

        self.assertEqual(s1.eval(x, 1), (1,))
        self.assertEqual(s2.eval(x, 1), (2,))

        sm1 = sm.branch()
        sm1.add(x == 1)
        self.assertEqual(sm1.eval(x, 1), (1,))
        self.assertEqual(sm1.eval(y, 1), (10,))
        self.assertEqual(sm1.eval(z, 1), (0,))
        self.assertEqual(sm1.eval(w, 1), (0,))

        sm2 = sm.branch()
        sm2.add(x == 2)
        self.assertEqual(sm2.eval(x, 1), (2,))
        self.assertEqual(sm2.eval(y, 1), (0,))
        self.assertEqual(sm2.eval(z, 1), (20,))
        self.assertEqual(sm2.eval(w, 1), (5,))

        sm1 = sm.branch()
        sm1.add(m == 0)
        self.assertEqual(sm1.eval(x, 1), (1,))
        self.assertEqual(sm1.eval(y, 1), (10,))
        self.assertEqual(sm1.eval(z, 1), (0,))
        self.assertEqual(sm1.eval(w, 1), (0,))

        sm2 = sm.branch()
        sm2.add(m == 1)
        self.assertEqual(sm2.eval(x, 1), (2,))
        self.assertEqual(sm2.eval(y, 1), (0,))
        self.assertEqual(sm2.eval(z, 1), (20,))
        self.assertEqual(sm2.eval(w, 1), (5,))

        m2 = claripy.BVS("m2", 32)
        _, smm = sm1.merge([sm2], [m2 == 0, m2 == 1])

        smm_1 = smm.branch()
        smm_1.add(x == 1)
        self.assertEqual(smm_1.eval(x, 1), (1,))
        self.assertEqual(smm_1.eval(y, 1), (10,))
        self.assertEqual(smm_1.eval(z, 1), (0,))
        self.assertEqual(smm_1.eval(w, 1), (0,))

        smm_2 = smm.branch()
        smm_2.add(m == 1)
        self.assertEqual(smm_2.eval(x, 1), (2,))
        self.assertEqual(smm_2.eval(y, 1), (0,))
        self.assertEqual(smm_2.eval(z, 1), (20,))
        self.assertEqual(smm_2.eval(w, 1), (5,))

        so = solver_type()
        so.add(w == 0)

        sa = so.branch()
        sb = so.branch()
        sa.add(x == 1)
        sb.add(x == 2)
        _, sm = sa.merge([sb], [m == 0, m == 1])

        smc = sm.branch()
        smd = sm.branch()
        smc.add(y == 3)
        smd.add(y == 4)

        _, smm = smc.merge([smd], [m2 == 0, m2 == 1])
        wxy = claripy.Concat(w, x, y)

        smm_1 = smm.branch()
        smm_1.add(wxy == 0x000103)
        self.assertTrue(smm_1.satisfiable())

        smm_1 = smm.branch()
        smm_1.add(wxy == 0x000104)
        self.assertTrue(smm_1.satisfiable())

        smm_1 = smm.branch()
        smm_1.add(wxy == 0x000203)
        self.assertTrue(smm_1.satisfiable())

        smm_1 = smm.branch()
        smm_1.add(wxy == 0x000204)
        self.assertTrue(smm_1.satisfiable())

        smm_1 = smm.branch()
        smm_1.add(wxy != 0x000103)
        smm_1.add(wxy != 0x000104)
        smm_1.add(wxy != 0x000203)
        smm_1.add(wxy != 0x000204)
        self.assertFalse(smm_1.satisfiable())

        s3 = solver_type()
        s3.add(w == 0)
        s4 = solver_type()
        _, sm = s3.merge([s4], [m == 0, m == 1])
        self.assertEqual(len(sm.eval(w, 2)), 2)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from .bool_check import is_false, is_true
from .ite_relocation import burrow_ite, excavate_ite
from .replace import replace, replace_dict
from .simplify import simplify

__all__ = (
    "burrow_ite",
    "excavate_ite",
    "is_false",
    "is_true",
    "replace",
    "replace_dict",
    "simplify",
)



from __future__ import annotations

import functools
import math
import struct
from decimal import Decimal

from claripy.backends.backend_object import BackendObject
from claripy.errors import ClaripyOperationError
from claripy.fp import FSORT_DOUBLE, FSORT_FLOAT, RM, FSort

from .bv import BVV, Concat


def compare_sorts(f):
    @functools.wraps(f)
    def compare_guard(self, o):
        if self.sort != o.sort:
            raise TypeError(f"FPVs are differently-sorted ({self.sort} and {o.sort})")
        return f(self, o)

    return compare_guard


def normalize_types(f):
    @functools.wraps(f)
    def normalize_helper(self, o):
        if isinstance(o, float):
            o = FPV(o, self.sort)

        if not isinstance(self, FPV) or not isinstance(o, FPV):
            raise TypeError("must have two FPVs")

        return f(self, o)

    return normalize_helper


class FPV(BackendObject):
    """A concrete floating point value. Used in the concrete backend for
    calculations.  Any use outside of claripy should use `claripy.FPV`
    instead.
    """

    __slots__ = ["sort", "value"]

    def __init__(self, value, sort):
        if not isinstance(value, float) or sort not in {FSORT_FLOAT, FSORT_DOUBLE}:
            raise ClaripyOperationError("FPV needs a sort (FSORT_FLOAT or FSORT_DOUBLE) and a float value")

        self.value = value
        self.sort = sort

    def __hash__(self):
        return hash((self.value, self.sort))

    def __getstate__(self):
        return self.value, self.sort

    def __setstate__(self, st):
        self.value, self.sort = st

    def __abs__(self):
        return FPV(abs(self.value), self.sort)

    def __neg__(self):
        return FPV(-self.value, self.sort)

    def fpSqrt(self):
        if self.value < 0:
            return FPV(float("nan"), self.sort)
        return FPV(math.sqrt(self.value), self.sort)

    @normalize_types
    @compare_sorts
    def __add__(self, o):
        return FPV(self.value + o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __sub__(self, o):
        return FPV(self.value - o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __mul__(self, o):
        return FPV(self.value * o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __mod__(self, o):
        return FPV(self.value % o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __truediv__(self, o):
        try:
            return FPV(self.value / o.value, self.sort)
        except ZeroDivisionError:
            if str(self.value * o.value)[0] == "-":
                return FPV(float("-inf"), self.sort)
            return FPV(float("inf"), self.sort)

    def __floordiv__(self, other):  # decline to involve integers in this floating point process
        return self.__truediv__(other)

    #
    # Reverse arithmetic stuff
    #

    @normalize_types
    @compare_sorts
    def __radd__(self, o):
        return FPV(o.value + self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rsub__(self, o):
        return FPV(o.value - self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rmul__(self, o):
        return FPV(o.value * self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rmod__(self, o):
        return FPV(o.value % self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rtruediv__(self, o):
        try:
            return FPV(o.value / self.value, self.sort)
        except ZeroDivisionError:
            if str(o.value * self.value)[0] == "-":
                return FPV(float("-inf"), self.sort)
            return FPV(float("inf"), self.sort)

    def __rfloordiv__(self, other):  # decline to involve integers in this floating point process
        return self.__rtruediv__(other)

    #
    # Boolean stuff
    #

    @normalize_types
    @compare_sorts
    def __eq__(self, o):
        return self.value == o.value

    @normalize_types
    @compare_sorts
    def __ne__(self, o):
        return self.value != o.value

    @normalize_types
    @compare_sorts
    def __lt__(self, o):
        return self.value < o.value

    @normalize_types
    @compare_sorts
    def __gt__(self, o):
        return self.value > o.value

    @normalize_types
    @compare_sorts
    def __le__(self, o):
        return self.value <= o.value

    @normalize_types
    @compare_sorts
    def __ge__(self, o):
        return self.value >= o.value

    def __repr__(self):
        return f"FPV({self.value:f}, {self.sort})"


def fpToFP(a1, a2, a3=None):
    """
    Returns a FP AST and has three signatures:

        fpToFP(ubvv, sort)
            Returns a FP AST whose value is the same as the unsigned BVV `a1`
            and whose sort is `a2`.

        fpToFP(rm, fpv, sort)
            Returns a FP AST whose value is the same as the floating point `a2`
            and whose sort is `a3`.

        fpToTP(rm, sbvv, sort)
            Returns a FP AST whose value is the same as the signed BVV `a2` and
            whose sort is `a3`.
    """
    if isinstance(a1, BVV) and isinstance(a2, FSort):
        sort = a2
        if sort == FSORT_FLOAT:
            pack, unpack = "I", "f"
        elif sort == FSORT_DOUBLE:
            pack, unpack = "Q", "d"
        else:
            raise ClaripyOperationError("unrecognized float sort")

        try:
            packed = struct.pack("<" + pack, a1.value)
            (unpacked,) = struct.unpack("<" + unpack, packed)
        except OverflowError as e:
            # struct.pack sometimes overflows
            raise ClaripyOperationError("OverflowError: " + str(e)) from e

        return FPV(unpacked, sort)
    if isinstance(a1, RM) and isinstance(a2, FPV) and isinstance(a3, FSort):
        return FPV(a2.value, a3)
    if isinstance(a1, RM) and isinstance(a2, BVV) and isinstance(a3, FSort):
        return FPV(float(a2.signed), a3)
    raise ClaripyOperationError("unknown types passed to fpToFP")


def fpToFPUnsigned(_rm, thing, sort):
    """
    Returns a FP AST whose value is the same as the unsigned BVV `thing` and
    whose sort is `sort`.
    """
    # thing is a BVV
    return FPV(float(thing.value), sort)


def fpToIEEEBV(fpv):
    """
    Interprets the bit-pattern of the IEEE754 floating point number `fpv` as a
    bitvector.

    :return:    A BV AST whose bit-pattern is the same as `fpv`
    """
    if fpv.sort == FSORT_FLOAT:
        pack, unpack = "f", "I"
    elif fpv.sort == FSORT_DOUBLE:
        pack, unpack = "d", "Q"
    else:
        raise ClaripyOperationError("unrecognized float sort")

    try:
        packed = struct.pack("<" + pack, fpv.value)
        (unpacked,) = struct.unpack("<" + unpack, packed)
    except OverflowError as e:
        # struct.pack sometimes overflows
        raise ClaripyOperationError("OverflowError: " + str(e)) from e

    return BVV(unpacked, fpv.sort.length)


def fpFP(sgn, exp, mantissa):
    """
    Concatenates the bitvectors `sgn`, `exp` and `mantissa` and returns the
    corresponding IEEE754 floating point number.

    :return:    A FP AST whose bit-pattern is the same as the concatenated
                bitvector
    """
    concatted = Concat(sgn, exp, mantissa)
    sort = FSort.from_size(concatted.size())

    if sort == FSORT_FLOAT:
        pack, unpack = "I", "f"
    elif sort == FSORT_DOUBLE:
        pack, unpack = "Q", "d"
    else:
        raise ClaripyOperationError("unrecognized float sort")

    try:
        packed = struct.pack("<" + pack, concatted.value)
        (unpacked,) = struct.unpack("<" + unpack, packed)
    except OverflowError as e:
        # struct.pack sometimes overflows
        raise ClaripyOperationError("OverflowError: " + str(e)) from e

    return FPV(unpacked, sort)


def fpToSBV(rm, fp, size):
    try:
        rounding_mode = rm.pydecimal_equivalent_rounding_mode()
        val = int(Decimal(fp.value).to_integral_value(rounding_mode))
        return BVV(val, size)

    except (ValueError, OverflowError):
        return BVV(0, size)
    except Exception as ex:
        print(f"Unhandled error during floating point rounding! {ex}")
        raise


def fpToUBV(rm, fp, size):
    # todo: actually make unsigned
    try:
        rounding_mode = rm.pydecimal_equivalent_rounding_mode()
        val = int(Decimal(fp.value).to_integral_value(rounding_mode))
        assert (
            val & ((1 << size) - 1) == val
        ), f"Rounding produced values outside the BV range! rounding {fp.value} with rounding mode {rm} produced {val}"
        if val < 0:
            val = (1 << size) + val
        return BVV(val, size)

    except (ValueError, OverflowError):
        return BVV(0, size)


def fpEQ(a, b):
    """
    Checks if floating point `a` is equal to floating point `b`.
    """
    return a == b


def fpNEQ(a, b):
    """
    Checks if floating point `a` is not equal to floating point `b`.
    """
    return a != b


def fpGT(a, b):
    """
    Checks if floating point `a` is greater than floating point `b`.
    """
    return a > b


def fpGEQ(a, b):
    """
    Checks if floating point `a` is greater than or equal to floating point `b`.
    """
    return a >= b


def fpLT(a, b):
    """
    Checks if floating point `a` is less than floating point `b`.
    """
    return a < b


def fpLEQ(a, b):
    """
    Checks if floating point `a` is less than or equal to floating point `b`.
    """
    return a <= b


def fpAbs(x):
    """
    Returns the absolute value of the floating point `x`. So:

        a = FPV(-3.2, FSORT_DOUBLE)
        b = fpAbs(a)
        b is FPV(3.2, FSORT_DOUBLE)
    """
    return abs(x)


def fpNeg(x):
    """
    Returns the additive inverse of the floating point `x`. So:

        a = FPV(3.2, FSORT_DOUBLE)
        b = fpAbs(a)
        b is FPV(-3.2, FSORT_DOUBLE)
    """
    return -x


def fpSub(_rm, a, b):
    """
    Returns the subtraction of the floating point `a` by the floating point `b`.
    """
    return a - b


def fpAdd(_rm, a, b):
    """
    Returns the addition of two floating point numbers, `a` and `b`.
    """
    return a + b


def fpMul(_rm, a, b):
    """
    Returns the multiplication of two floating point numbers, `a` and `b`.
    """
    return a * b


def fpDiv(_rm, a, b):
    """
    Returns the division of the floating point `a` by the floating point `b`.
    """
    return a / b


def fpIsNaN(x):
    """
    Checks whether the argument is a floating point NaN.
    """
    return math.isnan(x.value)


def fpIsInf(x):
    """
    Checks whether the argument is a floating point infinity.
    """
    return math.isinf(x.value)



# pylint:disable=duplicate-value,missing-class-docstring
from __future__ import annotations

import logging
import numbers
import operator
from functools import reduce

import claripy
from claripy.ast import BV, Bool
from claripy.backends.backend import Backend
from claripy.backends.backend_concrete import bv, fp, strings
from claripy.errors import BackendError, UnsatError
from claripy.operations import backend_fp_operations, backend_operations, backend_strings_operations

log = logging.getLogger(__name__)


# pylint: disable=too-many-positional-arguments


class BackendConcrete(Backend):
    __slots__ = ()

    def __init__(self):
        Backend.__init__(self)
        self._make_raw_ops(set(backend_operations) - {"If"}, op_module=bv)
        self._make_raw_ops(backend_strings_operations, op_module=strings)
        self._make_raw_ops(backend_fp_operations, op_module=fp)
        self._op_raw["If"] = self._If
        self._op_raw["BVV"] = self.BVV
        self._op_raw["StringV"] = self.StringV
        self._op_raw["FPV"] = self.FPV

        # reduceable
        self._op_raw["__add__"] = self._op_add
        self._op_raw["__sub__"] = self._op_sub
        self._op_raw["__mul__"] = self._op_mul
        self._op_raw["__or__"] = self._op_or
        self._op_raw["__xor__"] = self._op_xor
        self._op_raw["__and__"] = self._op_and

        # unary
        self._op_raw["__invert__"] = self._op_not
        self._op_raw["__neg__"] = self._op_neg
        self._op_raw["fpSqrt"] = self._op_fpSqrt

        # boolean ops
        self._op_raw["And"] = self._op_and
        self._op_raw["Or"] = self._op_or
        self._op_raw["Xor"] = self._op_xor
        self._op_raw["Not"] = self._op_boolnot

        self._cache_objects = False

    @staticmethod
    def BVV(value, size):
        return bv.BVV(value, size)

    @staticmethod
    def StringV(value):
        return strings.StringV(value)

    @staticmethod
    def FPV(op, sort):
        return fp.FPV(op, sort)

    @staticmethod
    def _op_add(*args):
        return reduce(operator.__add__, args)

    @staticmethod
    def _op_sub(*args):
        return reduce(operator.__sub__, args)

    @staticmethod
    def _op_mul(*args):
        return reduce(operator.__mul__, args)

    @staticmethod
    def _op_or(*args):
        return reduce(operator.__or__, args)

    @staticmethod
    def _op_xor(*args):
        return reduce(operator.__xor__, args)

    @staticmethod
    def _op_and(*args):
        return reduce(operator.__and__, args)

    @staticmethod
    def _op_not(arg):
        return ~arg

    @staticmethod
    def _op_neg(arg):
        return -arg

    @staticmethod
    def _op_boolnot(arg):
        return not arg

    @staticmethod
    def _op_fpSqrt(rm, a):  # pylint:disable=unused-argument
        return a.fpSqrt()

    def convert(self, expr):
        """
        Override Backend.convert() to add fast paths for BVVs and BoolVs.
        """
        if type(expr) is BV and expr.op == "BVV":
            cached_obj = self._object_cache.get(expr.hash(), None)
            if cached_obj is None:
                cached_obj = self.BVV(*expr.args)
                self._object_cache[expr.hash()] = cached_obj
            return cached_obj
        if type(expr) is Bool and expr.op == "BoolV":
            return expr.args[0]
        return super().convert(expr)

    def _If(self, b, t, f):  # pylint:disable=no-self-use,unused-argument
        if not isinstance(b, bool):
            raise BackendError("BackendConcrete can't handle non-bool condition in If.")
        return t if b else f

    def _name(self, o):  # pylint:disable=unused-argument,no-self-use
        return None

    def _identical(self, a, b):
        if type(a) is bv.BVV and type(b) is bv.BVV and a.size() != b.size():
            return False
        return a == b

    def _convert(self, r):
        if type(r) in {int, str, bytes}:
            return r
        if isinstance(r, numbers.Number | bv.BVV | fp.FPV | fp.RM | fp.FSort | strings.StringV):
            return r
        raise BackendError(f"can't handle AST of type {type(r)}")

    def _abstract(self, e):  # pylint:disable=no-self-use
        if isinstance(e, bv.BVV):
            return claripy.BVV(e.value, e.size())
        if isinstance(e, bool):
            return claripy.BoolV(e)
        if isinstance(e, fp.FPV):
            return claripy.FPV(e.value, e.sort)
        if isinstance(e, strings.StringV):
            return claripy.StringV(e.value)
        raise BackendError(f"Couldn't abstract object of type {type(e)}")

    def _cardinality(self, a):  # pylint:disable=unused-argument
        # if we got here, it's a cardinality of 1
        return 1

    #
    # Evaluation functions
    #

    @staticmethod
    def _to_primitive(expr):
        if isinstance(expr, bv.BVV | fp.FPV | strings.StringV):
            return expr.value
        if isinstance(expr, bool | numbers.Number):
            return expr
        raise BackendError("idk how to turn this into a primitive")

    def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")

        return (self._to_primitive(expr),)

    def _batch_eval(self, exprs, n, extra_constraints=(), solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")

        return [tuple(self._to_primitive(ex) for ex in exprs)]

    def _max(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")
        return self._to_primitive(expr)

    def _min(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")
        return self._to_primitive(expr)

    def _solution(self, expr, v, extra_constraints=(), solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")
        return self.convert(expr) == v

    # Override Backend.is_true() for a better performance
    def is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        if isinstance(e, numbers.Number):
            return bool(e)
        if e is claripy.true():
            return True
        return super().is_true(e, extra_constraints=extra_constraints, solver=solver, model_callback=model_callback)

    # Override Backend.is_false() for a better performance
    def is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        if isinstance(e, numbers.Number):
            return not bool(e)
        if e is claripy.false():
            return True
        return super().is_false(e, extra_constraints=extra_constraints, solver=solver, model_callback=model_callback)

    # pylint:disable=singleton-comparison
    def _is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == True  # noqa: E712

    def _is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == False  # noqa: E712

    def _has_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == True  # noqa: E712

    def _has_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == False  # noqa: E712



from __future__ import annotations

from .backend_concrete import BackendConcrete
from .bv import BVV
from .fp import FPV
from .strings import StringV

__all__ = ("BVV", "FPV", "BackendConcrete", "StringV")



from __future__ import annotations

import logging
import operator

import claripy
import claripy.backends.backend_vsa as vsa
from claripy.ast import BV, Base, Bool
from claripy.errors import BackendError, ClaripyBalancerError, ClaripyBalancerUnsatError, ClaripyOperationError
from claripy.operations import commutative_operations, opposites

log = logging.getLogger(__name__)


class Balancer:
    """
    The Balancer is an equation redistributor. The idea is to take an AST and rebalance it to, for example, isolate
    unknown terms on one side of an inequality.
    """

    def __init__(self, c):
        self._truisms = []
        self._ast_hash_map = {}
        self._lower_bounds = {}
        self._upper_bounds = {}

        self.sat = True
        try:
            self._doit(c)
        except ClaripyBalancerUnsatError:
            self.bounds = {}
            self.sat = False
        except BackendError:
            log.debug("Backend error in balancer.", exc_info=True)

    @property
    def compat_ret(self):
        return (self.sat, self.replacements)

    def _replacements_iter(self):
        all_keys = set(self._lower_bounds.keys()) | set(self._upper_bounds.keys())
        for k in all_keys:
            ast = self._ast_hash_map[k]
            max_int = (1 << len(ast)) - 1
            min_int = 0
            mn = self._lower_bounds.get(k, min_int)
            mx = self._upper_bounds.get(k, max_int)
            bound_si = claripy.BVS("bound", len(ast)).annotate(claripy.annotation.StridedIntervalAnnotation(1, mn, mx))
            log.debug("Yielding bound %s for %s.", bound_si, ast)
            if ast.op == "Reverse":
                yield (ast.args[0], ast.intersection(bound_si).reversed)
            else:
                yield (ast, ast.intersection(bound_si))

    def _add_lower_bound(self, o, b):
        if o.hash() in self._lower_bounds:
            old_b = self._lower_bounds[o.hash()]
            b = max(b, old_b)

        self._lower_bounds[o.hash()] = b
        self._ast_hash_map[o.hash()] = o

    def _add_upper_bound(self, o, b):
        if o.hash() in self._upper_bounds:
            old_b = self._upper_bounds[o.hash()]
            b = min(b, old_b)

        self._upper_bounds[o.hash()] = b
        self._ast_hash_map[o.hash()] = o

    @property
    def replacements(self):
        return list(self._replacements_iter())

    #
    # AST helper functions
    #

    @staticmethod
    def _same_bound_bv(a):
        si = claripy.backends.vsa.convert(a)
        mx = Balancer._max(a)
        mn = Balancer._min(a)
        return claripy.BVS("bounds", len(a)).annotate(claripy.annotation.StridedIntervalAnnotation(si._stride, mn, mx))

    @staticmethod
    def _cardinality(a):
        return a.cardinality if isinstance(a, Base) else 0

    @staticmethod
    def _min(a, signed=False):
        converted = claripy.backends.vsa.convert(a)
        if isinstance(converted, vsa.ValueSet):
            if len(converted.regions) == 1:
                converted = next(iter(converted.regions.values()))
            else:
                # unfortunately, this is a real abstract pointer
                # the minimum value will be 0 or MIN_INT
                if signed:
                    return -(1 << (len(converted) - 1))
                return 0
        bounds = converted._unsigned_bounds() if not signed else converted._signed_bounds()
        return min(mn for mn, mx in bounds)

    @staticmethod
    def _max(a, signed=False):
        converted = claripy.backends.vsa.convert(a)
        if isinstance(converted, vsa.ValueSet):
            if len(converted.regions) == 1:
                converted = next(iter(converted.regions.values()))
            else:
                # unfortunately, this is a real abstract pointer
                # the minimum value will be 0 or MIN_INT
                if signed:
                    return (1 << (len(converted) - 1)) - 1
                return (1 << len(converted)) - 1
        bounds = converted._unsigned_bounds() if not signed else converted._signed_bounds()
        return max(mx for mn, mx in bounds)

    @staticmethod
    def _range(a, signed=False):
        return (Balancer._min(a, signed=signed), Balancer._max(a, signed=signed))

    #
    # Truism alignment
    #

    @staticmethod
    def _align_truism(truism):
        outer_aligned = Balancer._align_ast(truism)
        inner_aligned = outer_aligned.make_like(
            outer_aligned.op, (Balancer._align_ast(outer_aligned.args[0]),) + outer_aligned.args[1:]
        )

        if not claripy.backends.vsa.identical(inner_aligned, truism):
            log.critical(
                "ERROR: the balancer is messing up an AST. This must be looked into. "
                "Please submit the binary and script to the angr project, if possible. "
                "Outer op is %s and inner op is %s.",
                truism.op,
                truism.args[0].op,
            )
            return truism

        return inner_aligned

    @staticmethod
    def _align_ast(a):
        """
        Aligns the AST so that the argument with the highest cardinality is on the left.

        :return: a new AST.
        """

        try:
            if isinstance(a, BV):
                return Balancer._align_bv(a)
            if isinstance(a, Bool) and len(a.args) == 2 and a.args[1].cardinality > a.args[0].cardinality:
                return Balancer._reverse_comparison(a)
            return a
        except ClaripyBalancerError:
            return a

    @staticmethod
    def _reverse_comparison(a):
        new_op = opposites.get(a.op, None)
        if new_op is None:
            raise ClaripyBalancerError(f"unable to reverse comparison {a.op} (missing from 'opposites')")

        try:
            op = getattr(operator, new_op)
        except AttributeError as err:
            raise ClaripyBalancerError(f"unable to reverse comparison {a.op} (AttributeError)") from err

        try:
            return op(*a.args[::-1])
        except ClaripyOperationError as err:
            raise ClaripyBalancerError(f"unable to reverse comparison {a.op} (ClaripyOperationError)") from err

    @staticmethod
    def _align_bv(a):
        if a.op in commutative_operations:
            return a.make_like(a.op, tuple(sorted(a.args, key=lambda v: -Balancer._cardinality(v))))

        match a.op:
            case "__sub__":
                return Balancer._align_sub(a)
            case _:
                return a

    @staticmethod
    def _align_sub(a):
        cardinalities = [Balancer._cardinality(v) for v in a.args]
        if max(cardinalities) == cardinalities[0]:
            return a

        adjusted = tuple(operator.__neg__(v) for v in a.args[1:]) + a.args[:1]
        return a.make_like("__add__", tuple(sorted(adjusted, key=lambda v: -Balancer._cardinality(v))))

    #
    # Find bounds
    #

    def _doit(self, c):
        """
        This function processes the list of truisms and finds bounds for ASTs.
        """
        self._truisms.append(claripy.excavate_ite(c))

        processed_truisms = set()
        identified_assumptions = set()

        while len(self._truisms):
            truism = self._truisms.pop()

            if truism in processed_truisms:
                continue

            unpacked_truisms = Balancer._unpack_truisms(truism)
            if claripy.backends.vsa.is_false(truism):
                raise ClaripyBalancerUnsatError

            processed_truisms.add(truism)
            if len(unpacked_truisms):
                self._truisms.extend(t for t in unpacked_truisms if not claripy.backends.vsa.is_true(t))
                continue

            if not Balancer._handleable_truism(truism):
                continue

            truism = Balancer._adjust_truism(truism)

            assumptions = Balancer._get_assumptions(truism)
            if truism not in identified_assumptions and len(assumptions):
                log.debug("Queued assumptions %s for truism %s.", assumptions, truism)
                self._truisms.extend(assumptions)
                identified_assumptions.update(assumptions)

            log.debug("Processing truism %s", truism)
            balanced_truism = self._balance(truism)
            log.debug("... handling")
            self._handle(balanced_truism)

    @staticmethod
    def _handleable_truism(t):
        """
        Checks whether we can handle this truism. The truism should already be aligned.
        """
        if len(t.args) < 2:
            log.debug("can't do anything with an unop bool")
            return None
        if t.args[0].cardinality > 1 and t.args[1].cardinality > 1:
            log.debug("can't do anything because we have multiple multivalued guys")
            return False
        if t.op == "If":
            log.debug("can't handle If")
            return False
        return True

    @staticmethod
    def _adjust_truism(t):
        """
        Swap the operands of the truism if the unknown variable is on the right side and the concrete value is on the
        left side.
        """
        if t.args[0].cardinality == 1 and t.args[1].cardinality > 1:
            return Balancer._reverse_comparison(t)
        return t

    #
    # Assumptions management
    #

    @staticmethod
    def _get_assumptions(t):
        """
        Given a constraint, _get_assumptions() returns a set of constraints that are implicitly
        assumed to be true. For example, `x <= 10` would return `x >= 0`.
        """

        if t.op in ("__le__", "__lt__", "ULE", "ULT"):
            return [t.args[0] >= 0]
        if t.op in ("__ge__", "__gt__", "UGE", "UGT"):
            return [t.args[0] <= 2 ** len(t.args[0]) - 1]
        if t.op in ("SLE", "SLT"):
            return [claripy.SGE(t.args[0], -(1 << (len(t.args[0]) - 1)))]
        if t.op in ("SGE", "SGT"):
            return [claripy.SLE(t.args[0], (1 << (len(t.args[0]) - 1)) - 1)]
        return []

    #
    # Truism extractor
    #

    @staticmethod
    def _unpack_truisms(c) -> set:
        """
        Given a constraint, _unpack_truisms() returns a set of constraints that must be True for
        this constraint to be True.
        """
        match c.op:
            case "And":
                return Balancer._unpack_truisms_and(c)
            case "Not":
                return Balancer._unpack_truisms_not(c)
            case "Or":
                return Balancer._unpack_truisms_or(c)
            case _:
                return set()

    @staticmethod
    def _unpack_truisms_and(c):
        return set.union(*[Balancer._unpack_truisms(a) for a in c.args])

    @staticmethod
    def _unpack_truisms_not(c):
        if c.args[0].op == "And":
            return Balancer._unpack_truisms(claripy.Or(*[claripy.Not(a) for a in c.args[0].args]))
        if c.args[0].op == "Or":
            return Balancer._unpack_truisms(claripy.And(*[claripy.Not(a) for a in c.args[0].args]))
        return set()

    @staticmethod
    def _unpack_truisms_or(c):
        vals = [claripy.backends.vsa.is_false(v) for v in c.args]
        if all(vals):
            raise ClaripyBalancerUnsatError
        if vals.count(False) == 1:
            return Balancer._unpack_truisms(c.args[vals.index(False)])
        return set()

    #
    # Simplification routines
    #

    def _balance(self, truism):
        while True:
            log.debug("Balancing %s", truism)

            # can't balance single-arg bools (Not) for now
            if len(truism.args) == 1:
                return truism

            if not isinstance(truism.args[0], Base):
                return truism

            try:
                inner_aligned = Balancer._align_truism(truism)
                if inner_aligned.args[1].cardinality > 1:
                    log.debug("can't do anything because we have multiple multivalued guys")
                    return truism

                match inner_aligned.args[0].op:
                    case "Reverse":
                        balanced = Balancer._balance_reverse(inner_aligned)
                    case "__add__":
                        balanced = Balancer._balance_add(inner_aligned)
                    case "__sub__":
                        balanced = Balancer._balance_sub(inner_aligned)
                    case "ZeroExt":
                        balanced = Balancer._balance_zeroext(inner_aligned)
                    case "SignExt":
                        balanced = Balancer._balance_signext(inner_aligned)
                    case "Extract":
                        balanced = Balancer._balance_extract(inner_aligned)
                    case "__and__":
                        balanced = Balancer._balance_and(inner_aligned)
                    case "Concat":
                        balanced = Balancer._balance_concat(inner_aligned)
                    case "__lshift__":
                        balanced = Balancer._balance_lshift(inner_aligned)
                    case "If":
                        balanced = self._balance_if(inner_aligned)
                    case _:
                        log.debug("Balance handler %s not implemented.", truism.args[0].op)
                        return truism

                if balanced is inner_aligned:
                    return balanced
                truism = balanced
                continue
            except ClaripyBalancerError:
                log.warning("Balance handler for operation %s raised exception.", truism.args[0].op)
                return truism

    @staticmethod
    def _balance_reverse(truism):
        if truism.op in ["__eq__", "__ne__"]:
            return truism.make_like(truism.op, (truism.args[0].args[0], truism.args[1].reversed))
        return truism

    @staticmethod
    def _balance_add(truism):
        if len(truism.args) != 2:
            return truism
        old_rhs = truism.args[1]
        lhs = truism.args[0]
        if all(a.concrete for a in lhs.args):
            # the old logic
            new_lhs = lhs.args[0]
            other_adds = lhs.args[1:]
        else:
            new_lhs = tuple(a for a in lhs.args if a.symbolic)
            if not new_lhs:
                return truism
            new_lhs = new_lhs[0] if len(new_lhs) == 1 else lhs.make_like("__add__", new_lhs)
            other_adds = tuple(a for a in lhs.args if a.concrete)
            if not other_adds:
                return truism
        new_rhs = truism.args[0].make_like("__sub__", (old_rhs, *other_adds))
        return truism.make_like(truism.op, (new_lhs, new_rhs))

    @staticmethod
    def _balance_sub(truism):
        if len(truism.args) != 2:
            return truism
        new_lhs = truism.args[0].args[0]
        old_rhs = truism.args[1]
        other_adds = truism.args[0].args[1:]
        new_rhs = truism.args[0].make_like("__add__", (old_rhs, *other_adds))
        return truism.make_like(truism.op, (new_lhs, new_rhs))

    @staticmethod
    def _balance_zeroext(truism):
        num_zeroes, inner = truism.args[0].args
        other_side = truism.args[1][len(truism.args[1]) - 1 : len(truism.args[1]) - num_zeroes]

        if claripy.backends.vsa.is_true(other_side == 0):
            # We can safely eliminate this layer of ZeroExt
            new_args = (inner, truism.args[1][len(truism.args[1]) - num_zeroes - 1 : 0])
            return truism.make_like(truism.op, new_args)

        return truism

    @staticmethod
    def _balance_signext(truism):
        num_zeroes = truism.args[0].args[0]
        left_side = truism.args[0][len(truism.args[1]) - 1 : len(truism.args[1]) - num_zeroes]
        other_side = truism.args[1][len(truism.args[1]) - 1 : len(truism.args[1]) - num_zeroes]

        # TODO: what if this is a set value, but *not* the same as other_side
        if claripy.backends.vsa.identical(left_side, other_side):
            # We can safely eliminate this layer of ZeroExt
            new_args = (truism.args[0].args[1], truism.args[1][len(truism.args[1]) - num_zeroes - 1 : 0])
            return truism.make_like(truism.op, new_args)

        return truism

    @staticmethod
    def _balance_extract(truism):
        high, low, inner = truism.args[0].args
        inner_size = len(inner)

        if high < inner_size - 1:
            left_msb = inner[inner_size - 1 : high + 1]
            left_msb_zero = claripy.backends.vsa.is_true(left_msb == 0)
        else:
            left_msb = None
            left_msb_zero = None

        if low > 0:
            left_lsb = inner[high - 1 : 0]
            left_lsb_zero = claripy.backends.vsa.is_true(left_lsb == 0)
        else:
            left_lsb = None
            left_lsb_zero = None

        if left_msb_zero and left_lsb_zero:
            new_left = inner
            new_right = claripy.Concat(claripy.BVV(0, len(left_msb)), truism.args[1], claripy.BVV(0, len(left_lsb)))
            return truism.make_like(truism.op, (new_left, new_right))
        if left_msb_zero:
            new_left = inner
            new_right = claripy.Concat(claripy.BVV(0, len(left_msb)), truism.args[1])
            return truism.make_like(truism.op, (new_left, new_right))
        if left_lsb_zero:
            new_left = inner
            new_right = claripy.Concat(truism.args[1], claripy.BVV(0, len(left_lsb)))
            return truism.make_like(truism.op, (new_left, new_right))

        if low == 0 and truism.args[1].op == "BVV" and truism.op not in {"SGE", "SLE", "SGT", "SLT"}:
            # single-valued rhs value with an unsigned operator
            # Eliminate Extract on lhs and zero-extend the value on rhs
            new_left = inner
            new_right = claripy.ZeroExt(inner.size() - truism.args[1].size(), truism.args[1])
            return truism.make_like(truism.op, (new_left, new_right))

        return truism

    @staticmethod
    def _balance_and(truism):
        if len(truism.args[0].args) != 2:
            return truism
        op0, op1 = truism.args[0].args

        if op1.op == "BVV":
            # if all low bits of right are 1 and all high bits of right are 0, then this is equivalent to Extract()
            v = op1.args[0]
            low_ones = 0
            while v != 0:
                if v & 1 == 0:
                    # not all high bits are 0. abort
                    return truism
                low_ones += 1
                v >>= 1
            if low_ones == 0:
                # this should probably never happen
                new_left = truism.args[0].make_like("BVV", (0, truism.args[0].size()))
                return truism.make_like(truism.op, (new_left, truism.args[1]))

            if op0.op == "ZeroExt" and op0.args[0] + low_ones == op0.size():
                # ZeroExt(56, a) & 0xff == a  if a.size() == 8
                # we can safely remove __and__
                new_left = op0
                return truism.make_like(truism.op, (new_left, truism.args[1]))

        return truism

    @staticmethod
    def _balance_concat(truism):
        size = len(truism.args[0])
        left_msb = truism.args[0].args[0]
        right_msb = truism.args[1][size - 1 : size - len(left_msb)]

        if claripy.backends.vsa.is_true(left_msb == 0) and claripy.backends.vsa.is_true(right_msb == 0):
            # we can cut these guys off!
            remaining_left = claripy.Concat(*truism.args[0].args[1:])
            remaining_right = truism.args[1][size - len(left_msb) - 1 : 0]
            return truism.make_like(truism.op, (remaining_left, remaining_right))
        # TODO: handle non-zero single-valued cases
        return truism

    @staticmethod
    def _balance_lshift(truism):
        lhs = truism.args[0]
        rhs = truism.args[1]
        shift_amount_expr = lhs.args[1]
        expr = lhs.args[0]

        shift_amount_values = claripy.backends.vsa.eval(shift_amount_expr, 2)
        if len(shift_amount_values) != 1:
            return truism
        shift_amount = shift_amount_values[0]

        rhs_lower = claripy.Extract(shift_amount - 1, 0, rhs)
        rhs_lower_values = claripy.backends.vsa.eval(rhs_lower, 2)
        if len(rhs_lower_values) == 1 and rhs_lower_values[0] == 0:
            # we can remove the __lshift__

            return truism.make_like(truism.op, (expr, rhs >> shift_amount))

        return truism

    def _balance_if(self, truism):
        condition, true_expr, false_expr = truism.args[0].args

        try:
            true_condition = getattr(true_expr, truism.op)(truism.args[1])
            false_condition = getattr(false_expr, truism.op)(truism.args[1])
        except ClaripyOperationError:
            # the condition was probably a Not (TODO)
            return truism

        can_true = claripy.backends.vsa.has_true(true_condition)
        can_false = claripy.backends.vsa.has_true(false_condition)
        must_true = claripy.backends.vsa.is_true(true_condition)
        must_false = claripy.backends.vsa.is_true(false_condition)

        if can_true and can_false:
            # always satisfiable
            return truism
        if not (can_true or can_false):
            # neither are satisfiable. This truism is fucked
            raise ClaripyBalancerUnsatError
        if must_true or (can_true and not can_false):
            # it will always be true
            self._truisms.append(condition)
            return truism.make_like(truism.op, (true_expr, truism.args[1]))
        if must_false or (can_false and not can_true):
            # it will always be false
            self._truisms.append(~condition)
            return truism.make_like(truism.op, (false_expr, truism.args[1]))
        return None

    #
    # Constraint handlers
    #

    def _handle(self, truism):
        log.debug("Handling %s", truism)

        if claripy.backends.vsa.is_false(truism):
            raise ClaripyBalancerUnsatError
        if Balancer._cardinality(truism.args[0]) == 1:
            # we are down to single-cardinality arguments, so our work is not
            # necessary
            return

        match truism.op:
            case "__eq__":
                self._handle_eq(truism)
            case "__ne__":
                self._handle_ne(truism)
            case "If":
                self._handle_if(truism)
            case (
                "__lt__"
                | "__le__"
                | "__gt__"
                | "__ge__"
                | "ULT"
                | "ULE"
                | "UGT"
                | "UGE"
                | "SLT"
                | "SLE"
                | "SGT"
                | "SGE"
            ):
                self._handle_comparison(truism)
            case _:
                log.debug("No handler for operation %s", truism.op)

    comparison_info = {  # noqa: RUF012
        "ULT": (True, False, True),
        "ULE": (True, True, True),
        "UGT": (False, False, True),
        "UGE": (False, True, True),
        "SLT": (True, False, False),
        "SLE": (True, True, False),
        "SGT": (False, False, False),
        "SGE": (False, True, False),
        "__lt__": (True, False, False),
        "__le__": (True, True, False),
        "__gt__": (False, False, False),
        "__ge__": (False, True, False),
    }

    def _handle_comparison(self, truism):
        """
        Handles all comparisons.
        """

        is_lt, is_equal, is_unsigned = self.comparison_info[truism.op]

        size = len(truism.args[0])
        int_max = 2**size - 1 if is_unsigned else 2 ** (size - 1) - 1
        int_min = -(2 ** (size - 1))

        left_min = Balancer._min(truism.args[0], signed=not is_unsigned)
        left_max = Balancer._max(truism.args[0], signed=not is_unsigned)
        right_min = Balancer._min(truism.args[1], signed=not is_unsigned)
        right_max = Balancer._max(truism.args[1], signed=not is_unsigned)

        bound_max = right_max if is_equal else (right_max - 1 if is_lt else right_max + 1)
        bound_min = right_min if is_equal else (right_min - 1 if is_lt else right_min + 1)

        if is_lt and bound_max < int_min:
            # if the bound max is negative and we're unsigned less than, we're fucked
            raise ClaripyBalancerUnsatError
        if not is_lt and bound_min > int_max:
            # if the bound min is too big, we're fucked
            raise ClaripyBalancerUnsatError

        current_min = int_min
        current_max = int_max

        if is_lt:
            current_max = min(int_max, left_max, bound_max)
            self._add_upper_bound(truism.args[0], current_max)
        else:
            current_min = max(int_min, left_min, bound_min)
            self._add_lower_bound(truism.args[0], current_min)

    def _handle_eq(self, truism):
        lhs, rhs = truism.args
        if rhs.cardinality != 1:
            common = Balancer._same_bound_bv(lhs.intersection(rhs))
            mn, mx = Balancer._range(common)
            self._add_upper_bound(lhs, mx)
            self._add_upper_bound(rhs, mx)
            self._add_lower_bound(lhs, mn)
            self._add_lower_bound(rhs, mn)
        else:
            mn, mx = Balancer._range(rhs)
            self._add_upper_bound(lhs, mx)
            self._add_lower_bound(lhs, mn)

    def _handle_ne(self, truism):
        lhs, rhs = truism.args
        if rhs.cardinality == 1:
            val = claripy.backends.vsa.eval(rhs, 1)[0]
            max_int = vsa.StridedInterval.max_int(len(rhs))

            if val == 0:
                self._add_lower_bound(lhs, val + 1)
            elif val in (max_int, val - 1):
                self._add_upper_bound(lhs, max_int - 1)

    def _handle_if(self, truism):
        if claripy.backends.vsa.is_false(truism.args[2]):
            self._truisms.append(truism.args[0])
        elif claripy.backends.vsa.is_false(truism.args[1]):
            self._truisms.append(~truism.args[0])



from __future__ import annotations

from claripy import algorithm, annotation, ast, backends
from claripy.algorithm import burrow_ite, excavate_ite, is_false, is_true, replace, replace_dict, simplify
from claripy.annotation import Annotation, RegionAnnotation, SimplificationAvoidanceAnnotation, UninitializedAnnotation
from claripy.ast.bool import (
    And,
    BoolS,
    BoolV,
    If,
    Not,
    Or,
    constraint_to_si,
    false,
    ite_cases,
    ite_dict,
    reverse_ite_cases,
    true,
)
from claripy.ast.bv import (
    BVS,
    BVV,
    ESI,
    SGE,
    SGT,
    SI,
    SLE,
    SLT,
    TSI,
    UGE,
    UGT,
    ULE,
    ULT,
    VS,
    Concat,
    Extract,
    LShR,
    Reverse,
    RotateLeft,
    RotateRight,
    SDiv,
    SignExt,
    SMod,
    ValueSet,
    ZeroExt,
    intersection,
    union,
    widen,
)
from claripy.ast.fp import (
    FPS,
    FPV,
    RM,
    fpAbs,
    fpAdd,
    fpDiv,
    fpEQ,
    fpFP,
    fpGEQ,
    fpGT,
    fpIsInf,
    fpIsNaN,
    fpLEQ,
    fpLT,
    fpMul,
    fpNeg,
    fpNEQ,
    fpSqrt,
    fpSub,
    fpToFP,
    fpToFPUnsigned,
    fpToIEEEBV,
    fpToSBV,
    fpToUBV,
)
from claripy.ast.strings import (
    IntToStr,
    StrConcat,
    StrContains,
    StrIndexOf,
    StringS,
    StringV,
    StrIsDigit,
    StrLen,
    StrPrefixOf,
    StrReplace,
    StrSubstr,
    StrSuffixOf,
    StrToInt,
)
from claripy.debug import set_debug
from claripy.errors import (
    ClaripyError,
    ClaripyFrontendError,
    ClaripyOperationError,
    ClaripySolverInterruptError,
    ClaripyZeroDivisionError,
    UnsatError,
)
from claripy.fp import FSORT_DOUBLE, FSORT_FLOAT
from claripy.solvers import (
    Solver,
    SolverCacheless,
    SolverComposite,
    SolverConcrete,
    SolverHybrid,
    SolverReplacement,
    SolverStrings,
    SolverVSA,
)

__version__ = "9.2.152"

__all__ = (
    "BVS",
    "BVV",
    "ESI",
    "FPS",
    "FPV",
    "FSORT_DOUBLE",
    "FSORT_FLOAT",
    "RM",
    "SGE",
    "SGT",
    "SI",
    "SLE",
    "SLT",
    "TSI",
    "UGE",
    "UGT",
    "ULE",
    "ULT",
    "VS",
    "And",
    "Annotation",
    "BoolS",
    "BoolV",
    "ClaripyError",
    "ClaripyFrontendError",
    "ClaripyOperationError",
    "ClaripySolverInterruptError",
    "ClaripyZeroDivisionError",
    "Concat",
    "Extract",
    "If",
    "IntToStr",
    "LShR",
    "Not",
    "Or",
    "RegionAnnotation",
    "Reverse",
    "RotateLeft",
    "RotateRight",
    "SDiv",
    "SMod",
    "SignExt",
    "SimplificationAvoidanceAnnotation",
    "Solver",
    "SolverCacheless",
    "SolverComposite",
    "SolverConcrete",
    "SolverHybrid",
    "SolverReplacement",
    "SolverStrings",
    "SolverVSA",
    "StrConcat",
    "StrContains",
    "StrIndexOf",
    "StrIsDigit",
    "StrLen",
    "StrPrefixOf",
    "StrReplace",
    "StrSubstr",
    "StrSuffixOf",
    "StrToInt",
    "StringS",
    "StringV",
    "UninitializedAnnotation",
    "UnsatError",
    "ValueSet",
    "ZeroExt",
    "algorithm",
    "annotation",
    "ast",
    "backends",
    "burrow_ite",
    "constraint_to_si",
    "excavate_ite",
    "false",
    "fpAbs",
    "fpAdd",
    "fpDiv",
    "fpEQ",
    "fpFP",
    "fpGEQ",
    "fpGT",
    "fpIsInf",
    "fpIsNaN",
    "fpLEQ",
    "fpLT",
    "fpMul",
    "fpNEQ",
    "fpNeg",
    "fpSqrt",
    "fpSub",
    "fpToFP",
    "fpToFPUnsigned",
    "fpToIEEEBV",
    "fpToSBV",
    "fpToUBV",
    "intersection",
    "is_false",
    "is_true",
    "ite_cases",
    "ite_dict",
    "replace",
    "replace_dict",
    "reverse_ite_cases",
    "set_debug",
    "simplify",
    "true",
    "union",
    "widen",
)



from __future__ import annotations

from claripy.backends.backend_object import BackendObject


class BoolResult(BackendObject):
    """A class representing the result of a boolean operation. Values can be
    True, False, or Maybe.
    """

    value: tuple[bool, ...]

    def __init__(self, value: tuple[bool, ...]):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, BoolResult) and self.value == other.value

    def __and__(self, other):
        if BoolResult.is_false(self) or BoolResult.is_false(other):
            return FalseResult()
        if BoolResult.is_true(self):
            return other
        if BoolResult.is_true(other):
            return self
        return MaybeResult()

    def __invert__(self):
        if BoolResult.is_true(self):
            return FalseResult()
        if BoolResult.is_false(self):
            return TrueResult()
        return MaybeResult()

    def __or__(self, other):
        if BoolResult.is_true(self) or BoolResult.is_true(other):
            return TrueResult()
        if BoolResult.is_false(self):
            return other
        if BoolResult.is_false(other):
            return self
        return MaybeResult()

    def identical(self, other):
        return self.value == other.value

    def union(self, other):
        return BoolResult(tuple(set(self.value) | set(other.value)))

    @property
    def cardinality(self):
        return len(self.value)

    @staticmethod
    def is_maybe(o):
        return isinstance(o, BoolResult) and False in o.value and True in o.value

    @staticmethod
    def has_true(o):
        return o is True or (isinstance(o, BoolResult) and True in o.value)

    @staticmethod
    def has_false(o):
        return o is False or (isinstance(o, BoolResult) and False in o.value)

    @staticmethod
    def is_true(o):
        return o is True or (isinstance(o, BoolResult) and o.value == (True,))

    @staticmethod
    def is_false(o):
        return o is False or (isinstance(o, BoolResult) and o.value == (False,))


def TrueResult() -> BoolResult:
    """Return a BoolResult representing the value True."""

    return BoolResult((True,))


def FalseResult() -> BoolResult:
    """Return a BoolResult representing the value False."""

    return BoolResult((False,))


def MaybeResult() -> BoolResult:
    """Return a BoolResult representing the value Maybe."""

    return BoolResult((True, False))



# pylint:disable=isinstance-second-argument-not-valid-type
from __future__ import annotations

import collections
import itertools
import operator
from functools import reduce
from typing import TYPE_CHECKING

import claripy
from claripy.fp import FSORT_DOUBLE, FSORT_FLOAT

if TYPE_CHECKING:
    from claripy.ast import Base

SIMPLE_OPS = ("Concat", "SignExt", "ZeroExt")

extract_distributable = {
    "__and__",
    "__rand__",
    "__or__",
    "__ror__",
    "__xor__",
    "__rxor__",
}

flattenable = {
    "__and__",
    "__or__",
    "__xor__",
    "__mul__",
    "And",
    "Or",
}


def _deduplicate_filter(args):
    seen = set()
    new_args = []
    for arg in args:
        key = arg.hash()
        if key in seen:
            continue
        seen.add(key)
        new_args.append(arg)
    return new_args


#
# The simplifiers.
#


def if_simplifier(cond, if_true, if_false):
    # NOTE: this is never called; simplifications are implemented inline in the If op. why?
    if cond.is_true():
        return if_true

    if cond.is_false():
        return if_false
    return None


def concat_simplifier(*args):
    if len(args) == 1:
        return args[0]

    orig_args = args
    args = list(args)
    # length = sum(arg.length for arg in args)
    simplified = False

    if any(a.symbolic for a in args):
        i = 1
        # here, we concatenate any consecutive concrete terms
        while i < len(args):
            previous = args[i - 1]
            current = args[i]

            if (
                not (previous.symbolic or current.symbolic)
                and claripy.backends.concrete.handles(previous)
                and claripy.backends.concrete.handles(current)
            ):
                concatted = claripy.Concat(previous, current)
                # If the concrete arguments to concat have non-relocatable annotations attached,
                # we may not be able to simplify the concrete concat. This check makes sure we don't
                # create a nested concat in that case.
                #
                # This is necessary to ensure that simplified is set correctly. If we don't check this here,
                # later the concat-of-concat will eliminate the newly introduced concat again, meaning we end
                # up with the same args that we started with. But because we only eliminate the concat later,
                # the simplified variable would still be set to True after this loop which is wrong.
                if concatted.op != "Concat":
                    args[i - 1 : i + 1] = (concatted,)
                    continue

            i += 1

        if len(args) < len(orig_args):
            simplified = True

    # here, we flatten any concats among the arguments and remove zero-length arguments
    i = 0
    while i < len(args):
        current = args[i]
        if current.length == 0:
            args.pop(i)
            simplified = True
        elif current.op == "Concat":
            simplified = True
            args[i : i + 1] = current.args
            i += len(current.args)
        else:
            i += 1

    # here, we consolidate any consecutive concats on extracts from the same variable
    i = 0
    prev_var = None
    prev_left = None
    prev_right = None
    while i < len(args):
        if args[i].op != "Extract":
            prev_var = None
            prev_left = None
            prev_right = None
            i += 1
        elif prev_var is args[i].args[2] and prev_right == args[i].args[0] + 1:
            prev_right = args[i].args[1]
            args[i - 1 : i + 1] = [claripy.Extract(prev_left, prev_right, prev_var)]
            simplified = True
        else:
            prev_left = args[i].args[0]
            prev_right = args[i].args[1]
            prev_var = args[i].args[2]
            i += 1

    # if any(a.op == 'Reverse' for a in args):
    # 	  simplified = True
    # 	  args = [a.reversed for a in args]

    if simplified:
        return claripy.Concat(*args)

    return None


def rshift_simplifier(val, shift):
    if (shift == 0).is_true():
        return val
    if val.op == "Concat" and (val.args[0] == 0).is_true() and (shift > val.size() - val.args[0].size()).is_true():
        return claripy.BVV(0, val.size())
    if val.op == "ZeroExt" and (shift > val.size() - val.args[0]).is_true():
        return claripy.BVV(0, val.size())
    return None


def lshr_simplifier(val, shift):
    if (shift == 0).is_true():
        return val
    if val.op == "Concat" and (val.args[0] == 0).is_true() and (shift > val.size() - val.args[0].size()).is_true():
        return claripy.BVV(0, val.size())
    if val.op == "ZeroExt" and (shift > val.size() - val.args[0]).is_true():
        return claripy.BVV(0, val.size())
    return None


def lshift_simplifier(val, shift):
    if (shift == 0).is_true():
        return val
    if val.op == "__lshift__":
        real_val, inner_shift = val.args
        return real_val << (inner_shift + shift)
    return None


def eq_simplifier(a, b):
    if a is b:
        return claripy.true()

    if isinstance(a, claripy.ast.Bool) and b is claripy.true():
        return a
    if isinstance(b, claripy.ast.Bool) and a is claripy.true():
        return b
    if isinstance(a, claripy.ast.Bool) and b is claripy.false():
        return claripy.Not(a)
    if isinstance(b, claripy.ast.Bool) and a is claripy.false():
        return claripy.Not(b)

    if a.op == "Reverse" and b.op == "Reverse":
        return a.args[0] == b.args[0]

    # simple canonicalization
    if a.op == "BVV" and b.op != "BVV":
        return b == a

    # expr - c1 == c2    ->   expr == c1 + c2
    if a.op == "__sub__" and len(a.args) == 2 and a.args[1].op == "BVV" and b.op == "BVV":
        return a.args[0] == a.args[1] + b

    # 1 ^ expr == 0     ->   expr == 1
    if a.op == "__xor__" and b.op == "BVV" and b.args[0] == 0 and len(a.args) == 2:
        if a.args[1].op == "BVV" and a.args[1].args[0] == 1:  # expr ^ 1 == 0
            return a.args[0] == 1
        if a.args[0].op == "BVV" and a.args[0].args[0] == 1:  # 1 ^ expr == 0
            return a.args[1] == 1

    # TODO: all these ==/!= might really slow things down...
    if a.op == "If":
        if a.args[1] is b and claripy.is_true(a.args[2] != b):
            # (If(c, x, y) == x, x != y) -> c
            return a.args[0]
        if a.args[2] is b and claripy.is_true(a.args[1] != b):
            # (If(c, x, y) == y, x != y) -> !c
            return claripy.Not(a.args[0])

    if b.op == "If":
        if b.args[1] is a and claripy.is_true(b.args[2] != b):
            # (x == If(c, x, y)) -> c
            return b.args[0]
        if b.args[2] is a and claripy.is_true(b.args[1] != a):
            # (y == If(c, x, y)) -> !c
            return claripy.Not(b.args[0])

    # Masking and comparing against a constant
    simp = and_mask_comparing_against_constant_simplifier(operator.__eq__, a, b)
    if simp is not None:
        return simp

    # ZeroExt/Concat and Extract and comparing against a constant
    simp = zeroext_extract_comparing_against_constant_simplifier(operator.__eq__, a, b)
    if simp is not None:
        return simp

    # ZeroExt/Concat and comparing against a constant
    simp = zeroext_comparing_against_simplifier(operator.__eq__, a, b)
    if simp is not None:
        return simp

    if (a.op in SIMPLE_OPS or b.op in SIMPLE_OPS) and a.length > 1 and a.length == b.length:
        for i in range(a.length):
            a_bit = a[i:i]
            if a_bit.symbolic:
                break

            b_bit = b[i:i]
            if b_bit.symbolic:
                break

            if claripy.is_false(a_bit == b_bit):
                return claripy.false()
        return None
    return None


def ne_simplifier(a, b):
    if a is b:
        return claripy.false()

    if a.op == "Reverse" and b.op == "Reverse":
        return a.args[0] != b.args[0]

    # simple canonicalization
    if a.op == "BVV" and b.op != "BVV":
        return b != a

    if a.op == "If":
        if a.args[2] is b and claripy.is_true(a.args[1] != b):
            # (If(c, x, y) == x, x != y) -> c
            return a.args[0]
        if a.args[1] is b and claripy.is_true(a.args[2] != b):
            # (If(c, x, y) == y, x != y) -> !c
            return claripy.Not(a.args[0])

    if b.op == "If":
        if b.args[2] is a and claripy.is_true(b.args[1] != a):
            # (x == If(c, x, y)) -> c
            return b.args[0]
        if b.args[1] is a and claripy.is_true(b.args[2] != a):
            # (y == If(c, x, y)) -> !c
            return claripy.Not(b.args[0])

    # 1 ^ expr != 0     ->   expr == 0
    if a.op == "__xor__" and b.op == "BVV" and b.args[0] == 0 and len(a.args) == 2:
        if a.args[1].op == "BVV" and a.args[1].args[0] == 1:
            return a.args[0] != 1
        if a.args[0].op == "BVV" and a.args[0].args[0] == 1:
            return a.args[1] != 1

    # Masking and comparing against a constant
    simp = and_mask_comparing_against_constant_simplifier(operator.__ne__, a, b)
    if simp is not None:
        return simp

    # ZeroExt/Concat and Extract and comparing against a constant
    simp = zeroext_extract_comparing_against_constant_simplifier(operator.__ne__, a, b)
    if simp is not None:
        return simp

    # ZeroExt/Concat and comparing against a constant
    simp = zeroext_comparing_against_simplifier(operator.__ne__, a, b)
    if simp is not None:
        return simp

    if (a.op == SIMPLE_OPS or b.op in SIMPLE_OPS) and a.length > 1 and a.length == b.length:
        for i in range(a.length):
            a_bit = a[i:i]
            if a_bit.symbolic:
                break

            b_bit = b[i:i]
            if b_bit.symbolic:
                break

            if claripy.is_true(a_bit != b_bit):
                return claripy.true()
        return None
    return None


def ge_simplifier(a, b):
    # ZeroExt/Concat and comparing against a constant
    simp = zeroext_comparing_against_simplifier(operator.__ge__, a, b)
    if simp is not None:
        return simp
    return None


def bv_reverse_simplifier(body):
    if body.op == "Reverse":
        # Reverse(Reverse(x)) ==> x
        return body.args[0]

    if body.length == 8:
        # Reverse(byte) ==> byte
        return body

    if body.op == "Concat":
        if all(a.op == "Extract" for a in body.args):
            first_ast = body.args[0].args[2]
            for i, a in enumerate(body.args):
                if not (first_ast is a.args[2] and a.args[0] == ((i + 1) * 8 - 1) and a.args[1] == i * 8):
                    break
            else:
                upper_bound = body.args[-1].args[0]
                if first_ast.length == upper_bound + 1:
                    return first_ast
                return first_ast[upper_bound:0]
        if all(a.length == 8 for a in body.args):
            return claripy.Concat(*body.args[::-1])

    if body.op == "Concat" and all(a.op == "Reverse" for a in body.args) and all(a.length % 8 == 0 for a in body.args):
        return claripy.Concat(*[a.args[0] for a in reversed(body.args)])

    if body.op == "Extract" and body.args[2].op == "Reverse":
        # Reverse(Extract(hi, lo, Reverse(x))) ==> Extract(bits-lo-1, bits-hi-1, x)
        # Holds only when (hi+1) and lo are multiples of 8 (or, multiples of bits_per_byte if we really want to
        # suppport cLEMENCy)
        hi, lo = body.args[0:2]
        if (hi + 1) % 8 == 0 and lo % 8 == 0:
            x = body.args[2].args[0]
            new_hi = x.size() - lo - 1
            new_lo = x.size() - hi - 1
            return claripy.Extract(new_hi, new_lo, x)
        return None
    return None


def boolean_and_simplifier(*args):
    if len(args) == 1:
        return args[0]

    new_args = [None] * len(args)
    ctr = 0
    for a in args:
        if a.op == "BoolV":
            if a.is_false():
                return claripy.false()
        else:
            new_args[ctr] = a
            ctr += 1
    new_args = new_args[:ctr]

    if not new_args:
        return claripy.true()

    if len(new_args) < len(args):
        return claripy.And(*new_args)

    # a >= c && a != c    ->   a>c
    if len(args) == 2:
        a = args[0]
        b = args[1]
        if (
            len(a.args) >= 2
            and len(b.args) >= 2
            and a.args[0] is b.args[0]
            and a.args[1] is b.args[1]
            and a.op == "__ge__"
            and b.op == "__ne__"
        ):
            return claripy.UGT(a.args[0], a.args[1])

    flattened = _flatten_simplifier("And", _deduplicate_filter, *args)
    if flattened is None:
        return None

    if flattened.op != "And":
        return flattened

    fargs = flattened.args

    # Check if we are left with one argument again
    if len(fargs) == 1:
        return fargs[0]

    # what the fuck is this supposed to do?
    if any(len(arg.args) != 2 for arg in fargs):
        return flattened

    target_var = None

    # Determine the unknown variable
    if fargs[0].args[0].symbolic and (fargs[0].args[0] is fargs[1].args[0] or fargs[0].args[0] is fargs[1].args[1]):
        target_var = fargs[0].args[0]
    elif fargs[0].args[1].symbolic and (fargs[0].args[1] is fargs[1].args[0] or fargs[0].args[1] is fargs[1].args[1]):
        target_var = fargs[0].args[1]

    if target_var is None:
        return flattened

    # we now know that the And is a series of binary conditions over a single variable.
    # we can optimize accordingly.
    # right now it's just check for eq/ne

    eq_list = []
    ne_list = []
    for arg in fargs:
        other = arg.args[1] if arg.args[0] is target_var else arg.args[0] if arg.args[1] is target_var else None
        if other is None:
            return flattened
        if arg.op == "__eq__":
            eq_list.append(other)
        elif arg.op == "__ne__":
            ne_list.append(other)
        else:
            return flattened

    if not eq_list:
        return flattened
    if any(any(ne is eq for eq in eq_list) for ne in ne_list):
        return claripy.false()
    if all(v.op == "BVV" for v in eq_list) and all(v.op == "BVV" for v in ne_list):
        mustbe = eq_list[0]
        if any(eq.args[0] != mustbe.args[0] for eq in eq_list):
            return claripy.false()
        return target_var == eq_list[0]
    return flattened


def boolean_or_simplifier(*args):
    if len(args) == 1:
        return args[0]

    new_args = []
    for a in args:
        if a.is_true():
            return claripy.true()
        if not a.is_false():
            new_args.append(a)

    if not new_args:
        return claripy.false()
    if len(new_args) < len(args):
        return claripy.Or(*new_args)

    return _flatten_simplifier("Or", _deduplicate_filter, *args)


def _flatten_simplifier(op_name, filter_func, *args, **kwargs):
    # we cannot further flatten if any top-level argument has non-relocatable annotations
    if any(not anno.relocatable for anno in itertools.chain.from_iterable(arg.annotations for arg in args)):
        return None

    new_args = tuple(
        itertools.chain.from_iterable(
            (a.args if isinstance(a, claripy.ast.Base) and a.op == op_name and len(a.args) < 1000 else (a,))
            for a in args
        )
    )

    # try to collapse multiple concrete args
    value_args = []
    other_args = []
    for arg in new_args:
        if getattr(arg, "op", None) == "BVV":
            value_args.append(arg)
        else:
            other_args.append(arg)
    if other_args and len(value_args) > 1:
        value_arg = value_args[0].make_like(op_name, tuple(value_args), simplify=False)
        new_args = (*tuple(other_args), value_arg)

    variables = frozenset(itertools.chain.from_iterable(a.variables for a in args if isinstance(a, claripy.ast.Base)))
    if filter_func:
        new_args = filter_func(new_args)
    if not new_args and "initial_value" in kwargs:
        return kwargs["initial_value"]
    # if a single arg is left, don't create an op for it
    if len(new_args) == 1:
        return new_args[0]
    return next(a for a in args if isinstance(a, claripy.ast.Base)).make_like(
        op_name, new_args, variables=variables, simplify=False
    )


def bitwise_add_simplifier(*args):
    if len(args) == 2 and args[1].op == "BVV" and args[0].op == "__sub__" and args[0].args[1].op == "BVV":
        # flatten add over sub
        # (x - y) + z ==> x - (y - z)
        return args[0].args[0] - (args[0].args[1] - args[1])
    return _flatten_simplifier(
        "__add__",
        lambda new_args: tuple(a for a in new_args if a.op != "BVV" or a.args[0] != 0),
        *args,
        initial_value=claripy.BVV(0, len(args[0])),
    )


def bitwise_mul_simplifier(*args):
    return _flatten_simplifier("__mul__", None, *args)


def bitwise_sub_simplifier(a, b):
    if b.op == "BVV":
        # many optimizations if b is concrete - effectively flattening
        if b.args[0] == 0:
            return a
        if a.op == "__sub__" and a.args[1].op == "BVV":
            # flatten right-heavy trees
            # (x - y) - z ==> x - (y + z)
            return a.args[0] - (a.args[1] + b)
        if a.op == "__add__" and a.args[-1].op == "BVV":
            # flatten sub over add
            # (x + y) - z ==> x + (y - z)
            if len(a.args) == 2:
                return a.args[0] + (a.args[-1] - b)
            return a.make_like(a.op, (*a.args[:-1], a.args[-1] - b))
    elif a is b or (a == b).is_true():
        return claripy.BVV(0, a.size())
    return None


# recognize b-bit z=signedmax(q,r) from this idiom:
# s=q-r;t=q^r;u=s^q;v=u&t;w=v^s;x=rshift(w,b-1);y=x&t;z=q^y
# and recognize b-bit z=signedmin(q,r) from this idiom:
# s=r-q;t=q^r;u=s^r;v=u&t;w=v^s;x=rshift(w,b-1);y=x&t;z=q^y
def bitwise_xor_simplifier_minmax(a, b):
    q, y = a, b
    if y.op != "__and__":
        q, y = b, a
        if y.op != "__and__":
            return None

    if len(y.args) != 2:
        return None
    x, t = y.args
    if t.op != "__xor__":
        t, x = y.args
        if t.op != "__xor__":
            return None

    if x.op != "__rshift__":
        return None
    w, dist = x.args

    bits = a.size()
    if dist is not claripy.BVV(bits - 1, bits):
        return None
    if w.op != "__xor__":
        return None

    if len(w.args) != 2:
        return None
    v, s = w.args
    if s.op != "__sub__":
        s, v = w.args
        if s.op != "__sub__":
            return None

    if v.op != "__and__":
        return None
    if len(v.args) != 2:
        return None
    u, t2 = v.args
    if t2 is not t:
        t2, u = v.args
        if t2 is not t:
            return None

    if u.op != "__xor__":
        return None

    if len(t.args) != 2:
        return None
    q2, r = t.args
    if q2 is not q:
        r, q2 = t.args
        if q2 is not q:
            return None

    if (u.args[0] is s and u.args[1] is q) or (u.args[0] is q and u.args[1] is s):
        if not (s.args[0] is q and s.args[1] is r):
            return None
        cond = claripy.SLE(q, r)
        return claripy.If(cond, r, q)

    if (u.args[0] is s and u.args[1] is r) or (u.args[0] is r and u.args[1] is s):
        if not (s.args[0] is r and s.args[1] is q):
            return None
        cond = claripy.SLE(q, r)
        return claripy.If(cond, q, r)
    return None


def bitwise_xor_simplifier(a, b, *args):
    if not args:
        if a is claripy.BVV(0, a.size()):
            return b
        if b is claripy.BVV(0, a.size()):
            return a
        if a is b or (a == b).is_true():
            return claripy.BVV(0, a.size())

        result = bitwise_xor_simplifier_minmax(a, b)
        if result is not None:
            return result

    def _flattening_filter(args):
        # since a ^ a == 0, we can safely remove those from args
        # this procedure is done carefully in order to keep the ordering of arguments
        ctr = collections.Counter(arg.hash() for arg in args)
        res = []
        seen = set()
        for arg in args:
            if ctr[arg.hash()] % 2 == 0:
                continue
            l1 = len(seen)
            seen.add(arg.hash())
            l2 = len(seen)
            if l1 != l2:
                res.append(arg)
        return tuple(res)

    return _flatten_simplifier(
        "__xor__",
        _flattening_filter,
        a,
        b,
        *args,
        initial_value=claripy.BVV(0, a.size()),
    )


def bitwise_or_simplifier(a, b, *args):
    if not args:
        if a is claripy.BVV(0, a.size()):
            return b
        if b is claripy.BVV(0, a.size()):
            return a
        if a.op == b.op and a.op in {"BVV", "BoolV", "FPV"}:
            if a.args == b.args and (a == b).is_true():
                return a
        elif (a == b).is_true():
            return a
        if a is b:
            return a

    return _flatten_simplifier("__or__", _deduplicate_filter, a, b, *args)


def bitwise_and_simplifier(a, b, *args):
    if not args:
        # try to perform a rotate-shift-mask simplification
        r = rotate_shift_mask_simplifier(a, b)
        if r is not None:
            return r
        # we do not use (a == 2 ** a.size()-1).is_true() to avoid creating redundant claripys
        if a.op == "BVV" and a.args[0] == 2 ** a.size() - 1:
            return b
        if b.op == "BVV" and b.args[0] == 2 ** b.size() - 1:
            return a
        if a is b:
            return a
        # for concrete values, we delay the claripy creation as much as possible
        if a.op == b.op and a.op in {"BVV", "BoolV", "FPV"}:
            if a.args == b.args and (a == b).is_true():
                return a
        elif (a == b).is_true():
            return a
        if a.op == "BVV" and a.args[0] == 0:
            return claripy.BVV(0, a.size())
        if b.op == "BVV" and b.args[0] == 0:
            return claripy.BVV(0, a.size())
        # Concat(a.args[0], a.args[1]) & b  ==>  ZeroExt(size, a.args[1])
        # maybe we can drop the second argument
        if a.op == "Concat" and len(a.args) == 2 and (b == 2 ** (a.size() - a.args[0].size()) - 1).is_true():
            # yes!
            return claripy.ZeroExt(a.args[0].size(), a.args[1])

        # if(cond0, 1, 0) & if(cond1, 1, 0)  ->  if(cond0 & cond1, 1, 0)
        if (
            a.op == "If"
            and b.op == "If"
            and (
                (a.args[1] == claripy.BVV(1, len(a.args[1]))).is_true()
                and (a.args[2] == claripy.BVV(0, len(a.args[2]))).is_true()
                and (b.args[1] == claripy.BVV(1, len(a.args[1]))).is_true()
                and (b.args[2] == claripy.BVV(0, len(a.args[2]))).is_true()
            )
        ):
            cond0 = a.args[0]
            cond1 = b.args[0]
            return claripy.If(
                cond0 & cond1,
                claripy.BVV(1, len(a.args[1])),
                claripy.BVV(0, len(a.args[2])),
            )

    return _flatten_simplifier("__and__", _deduplicate_filter, a, b, *args)


def boolean_not_simplifier(body):
    if body.op == "__eq__":
        return body.args[0] != body.args[1]
    if body.op == "__ne__":
        return body.args[0] == body.args[1]

    if body.op == "Not":
        return body.args[0]

    if body.op == "SLT":
        return claripy.SGE(body.args[0], body.args[1])
    if body.op == "SLE":
        return claripy.SGT(body.args[0], body.args[1])
    if body.op == "SGT":
        return claripy.SLE(body.args[0], body.args[1])
    if body.op == "SGE":
        return claripy.SLT(body.args[0], body.args[1])

    if body.op == "ULT":
        return claripy.UGE(body.args[0], body.args[1])
    if body.op == "ULE":
        return claripy.UGT(body.args[0], body.args[1])
    if body.op == "UGT":
        return claripy.ULE(body.args[0], body.args[1])
    if body.op == "UGE":
        return claripy.ULT(body.args[0], body.args[1])

    if body.op == "__lt__":
        return claripy.UGE(body.args[0], body.args[1])
    if body.op == "__le__":
        return claripy.UGT(body.args[0], body.args[1])
    if body.op == "__gt__":
        return claripy.ULE(body.args[0], body.args[1])
    if body.op == "__ge__":
        return claripy.ULT(body.args[0], body.args[1])
    return None


def zeroext_simplifier(n, e):
    if n == 0:
        return e

    if e.op == "ZeroExt":
        # ZeroExt(A, ZeroExt(B, x)) ==> ZeroExt(A + B, x)
        claripy.ZeroExt(n + e.args[0], e.args[1])
    return None


def signext_simplifier(n, e):
    if n == 0:
        return e
    return None

    # TODO: if top bit is 0, do a zero-extend instead


def extract_simplifier(high, low, val):
    # if we're extracting the whole value, return the value
    if high - low + 1 == val.size():
        return val

    if val.op in {"SignExt", "ZeroExt"} and low == 0 and high + 1 == val.args[1].size():
        return val.args[1]

    if val.op == "ZeroExt":
        extending_bits = val.args[0]
        val = val.args[1] if extending_bits == 0 else claripy.Concat(claripy.BVV(0, extending_bits), val.args[1])

    # Reverse(concat(a, b)) -> concat(Reverse(b), Reverse(a))
    # a and b must have lengths that are a multiple of 8
    if val.op == "Reverse" and val.args[0].op == "Concat" and all(a.length % 8 == 0 for a in val.args[0].args):
        val = claripy.Concat(*reversed([a.reversed for a in val.args[0].args]))

    # Reading one byte from a reversed claripy can be converted to reading
    # the corresponding byte from the original claripy
    # No Reverse is required then
    if val.op == "Reverse" and high - low + 1 == 8 and low % 8 == 0:
        byte_pos = low // 8
        new_byte_pos = val.length // 8 - byte_pos - 1

        val = val.args[0]
        high = (new_byte_pos + 1) * 8 - 1
        low = new_byte_pos * 8

        return claripy.Extract(high, low, val)

    if val.op == "Concat":
        pos = val.length
        high_i, low_i, high_loc, low_loc = None, None, None, None
        for i, v in enumerate(val.args):
            if pos - v.length <= high < pos:
                high_i = i
                high_loc = high - (pos - v.length)
            if pos - v.length <= low < pos:
                low_i = i
                low_loc = low - (pos - v.length)
            pos -= v.length

        used = list(val.args[high_i : low_i + 1])

        # if we have only one part, we can avoid creating a new concat
        if len(used) == 1:
            new_high = low_loc + high - low
            result = used[0]

            # avoid extracting if we need the full value
            if new_high == result.length - 1 and low_loc == 0:
                return result, True

            # else extract the part we need
            return result[new_high:low_loc]

        # if we have more than one part, cut the parts that we need from the start & end chunks
        # if necessary
        if high_loc != used[0].length:
            used[0] = used[0][high_loc:]

        if low_loc != 0:
            used[-1] = used[-1][:low_loc]

        return claripy.Concat(*used)

    if val.op == "Extract":
        _, inner_low = val.args[:2]
        new_low = inner_low + low
        new_high = new_low + (high - low)
        return (val.args[2])[new_high:new_low]

    if val.op == "Reverse" and val.args[0].op == "Concat" and all(a.length % 8 == 0 for a in val.args[0].args):
        val = claripy.Concat(*reversed([a.reversed for a in val.args[0].args]))[high:low]
        if not val.symbolic:
            return val

    # if all else fails, convert Extract(Reverse(...)) to Reverse(Extract(...))
    # if val.op == 'Reverse' and (high + 1) % 8 == 0 and low % 8 == 0:
    # 	  print("saw reverse, converting")
    # 	  inner_length = val.args[0].length
    # 	  try:
    # 		  return val.args[0][(inner_length - 1 - low):(inner_length - 1 - low - (high - low))].reversed
    # 	  except ClaripyOperationError:
    # 		  __import__('ipdb').set_trace()

    if val.op in extract_distributable:
        all_args = tuple(a[high:low] for a in val.args)
        if val.op in flattenable:
            # directly create a flattened claripy
            return next(a for a in all_args if isinstance(a, claripy.ast.Base)).make_like(val.op, all_args)
        return reduce(getattr(operator, val.op), all_args)

    #  (if cond then 1 else 0)[0:0]  ->  if(cond then 1[0:0] else 0[0:0])
    if val.op == "If":
        ifcond, iftrue, iffalse = val.args
        if iftrue.op == "BVV" and iffalse.op == "BVV":
            # extract from iftrue and iffalse
            return claripy.If(ifcond, iftrue[high:low], iffalse[high:low])

    # invert(expr)[0:0]   ->   invert(expr[0:0])
    if val.op == "__invert__":
        return claripy.ast.BV.__invert__(val.args[0][high:low])
    return None


# oh gods
def fptobv_simplifier(the_fp):
    if the_fp.op == "fpToFP" and len(the_fp.args) == 2:
        return the_fp.args[0]
    return None


def fptofp_simplifier(*args):
    if len(args) == 2 and args[0].op == "fpToIEEEBV":
        to_bv, sort = args
        if (sort == FSORT_FLOAT and to_bv.length == 32) or (sort == FSORT_DOUBLE and to_bv.length == 64):
            return to_bv.args[0]
        return None
    return None


def rotate_shift_mask_simplifier(a, b):
    """
    Handles the following case:
        ((A << a) | (A >> (_N - a))) & mask, where
            A being a BVS,
            a being a integer that is less than _N,
            _N is either 32 or 64, and
            mask can be evaluated to 0xffffffff (64-bit) or 0xffff (32-bit) after reversing the rotate-shift
            operation.

    It will be simplified to:
        (A & (mask >>> a)) <<< a
    """

    # is the second argument a BVV?
    if b.op != "BVV":
        return None

    # is it a rotate-shift?
    if a.op != "__or__" or len(a.args) != 2:
        return None
    a_0, a_1 = a.args

    if a_0.op != "__lshift__":
        return None
    if a_1.op != "LShR":
        return None
    a_00, a_01 = a_0.args
    a_10, a_11 = a_1.args
    if a_00 is not a_10:
        return None
    if a_01.op != "BVV" or a_11.op != "BVV":
        return None
    lshift_ = a_01.args[0]
    rshift_ = a_11.args[0]
    bitwidth = lshift_ + rshift_
    if bitwidth not in (32, 64):
        return None

    # is the second argument a mask?
    # Note: the following check can be further loosen if we want to support more masks.
    if bitwidth == 32:
        m = ((b.args[0] << rshift_) & 0xFFFFFFFF) | (b.args[0] >> lshift_)
        if m != 0xFFFF:
            return None
    else:  # bitwidth == 64
        m = ((b.args[0] << rshift_) & 0xFFFFFFFFFFFFFFFF) | (b.args[0] >> lshift_)
        if m != 0xFFFFFFFF:
            return None

    # Show our power!
    masked_a = a_00 & m
    return (masked_a << lshift_) | (masked_a >> rshift_)


def str_reverse_simplifier(arg):
    return arg


def invert_simplifier(expr):
    # ~ if(cond then 1 else 0)  ->  if(cond, ~1, ~0)  ->    if(!cond, 1,0)
    if expr.op == "If" and expr.args[1].op == "BVV" and expr.args[1].args[0] == 1 and expr.args[2].args[0] == 0:
        return claripy.If(claripy.Not(expr.args[0]), expr.args[1], expr.args[2])
    return None


def and_mask_comparing_against_constant_simplifier(op, a, b):
    """
    This simplifier handles the following case:

        A & mask == b, and
        A & mask != b

    If the high bits of A are 0, `& mask` can be eliminated.
    """
    if op in (operator.__eq__, operator.__ne__) and a.op == "__and__" and len(a.args) == 2 and b.op == "BVV":
        a_arg0, a_arg1 = a.args
        if a_arg1.op == "BVV":
            # it's a constant mask
            # check if the higher bits are 0
            v = a_arg1.args[0]
            zero_bits = a_arg1.args[1]
            mask_allones = True
            while v != 0:
                mask_allones = mask_allones and ((v & 1) == 0)
                v = v >> 1
                zero_bits -= 1
            if zero_bits > 0:
                # the high `zero_bits` bits are zero
                # check the constant
                b_higher = b[b.size() - 1 : b.size() - zero_bits]
                b_higher_bits_are_0: bool | None = None
                if (b_higher == 0).is_true():
                    b_higher_bits_are_0 = True
                elif (b_higher == 0).is_false():
                    b_higher_bits_are_0 = False

                if b_higher_bits_are_0 is True:
                    # extra check: can we get rid of the mask
                    if zero_bits == b.size():
                        # the mask is 0, which means the left-hand side becomes 0 after masking
                        return op(claripy.BVV(0, b.size()), b)

                    b_highbit_idx = b.size() - 1 - zero_bits
                    # originally, b was 8-bit aligned. Can we keep the size of the new expression 8-byte aligned?
                    if b.size() % 8 == 0 and (b_highbit_idx + 1) % 8 != 0:
                        b_highbit_idx += 8 - (b_highbit_idx + 1) % 8
                    b_lower = b[b_highbit_idx:0]
                    if mask_allones:
                        # yes!
                        return op(a_arg0[b_highbit_idx:0], b_lower)
                    # nope
                    if b_highbit_idx == b.size() - 1:
                        # no change is happening
                        return None
                    return op(a_arg0[b_highbit_idx:0] & a_arg1.args[0], b_lower)
                if b_higher_bits_are_0 is False:
                    return claripy.false() if op is operator.__eq__ else claripy.true()

    return None


def zeroext_extract_comparing_against_constant_simplifier(op, a, b):
    """
    This simplifier handles the following cases:

        Extract(hi, 0, Concat(0, A)) op b, and
        Extract(hi, 0, ZeroExt(n, A)) op b

    Extract can be eliminated if the high bits of Concat(0, A) or ZeroExt(n, A) are all zeros.
    """

    # TODO: This method is only used when op is __eq__ or __ne__. We need more experiment to see if it is necessary
    # TODO: to apply this optimization to other comparison operators.

    if a.op != "Extract" or a.args[1] != 0:
        return None

    highbits_are_zeros = False

    a_hi, _, a_arg = a.args
    a_inner_expr = None
    if a_arg.op == "Concat" and len(a_arg.args) == 2:
        a_concat_expr0 = a_arg.args[0]
        a_inner_expr = a_arg.args[1]
        # equivalent to ZeroExt
        if a_concat_expr0.op == "BVV" and a_concat_expr0.args[0] == 0 and a_concat_expr0.size() >= a_arg.size() - a_hi:
            # high bits are all zeros
            highbits_are_zeros = True

    elif a_arg.op == "ZeroExt":
        a_zeroext_bits = a_arg.args[0]
        a_inner_expr = a_arg.args[1]
        if a_zeroext_bits >= a_arg.size() - a_hi:
            # high bits are all zeros
            highbits_are_zeros = True

    if highbits_are_zeros:
        to_extend = b.size() - a_inner_expr.size()
        if to_extend == 0:
            return op(a_inner_expr, b)
        return op(claripy.ZeroExt(to_extend, a_inner_expr), b)

    return None


def zeroext_comparing_against_simplifier(op, a, b):
    """
    This simplifier handles the following cases:

        ZeroExt(n, A) == b,
        ZeroExt(n, A) != b, and
        ZeroExt(n, A) >= b

    If the high bits of b are all zeros (in case of ==, !=, and >=) or have at leclaripy one ones (in case of !=),
    ZeroExt can be eliminated.
    """
    if op in {operator.__eq__, operator.__ne__, operator.__ge__} and b.op == "BVV":
        b_highbits = None
        removable_high_bits = 0
        if a.op == "ZeroExt":
            # check if the high bits of b are zeros
            a_zeroext_bits = a.args[0]
            b_highbits = b[b.size() - 1 : b.size() - a_zeroext_bits]
            removable_high_bits = a_zeroext_bits
        elif (
            a.op == "Concat" and len(a.args) == 2 and a.args[0].op == "BVV" and a.args[0].args[0] == 0
        ):  # make sure high bits of a are zeros
            a_zero_bits = a.args[0].args[1]
            b_highbits = b[b.size() - 1 : b.size() - a_zero_bits]
            removable_high_bits = a_zero_bits
        else:
            return None

        if (b_highbits == 0).is_true():
            # we can get rid of ZeroExt
            return op(a.args[1], b[b.size() - removable_high_bits - 1 : 0])
        if (b_highbits == 0).is_false():
            if op is operator.__ne__:
                return claripy.true()
            # op is __eq__, __ge__
            return claripy.false()

    return None


_all_simplifiers = {
    "And": boolean_and_simplifier,
    "Or": boolean_or_simplifier,
    "Not": boolean_not_simplifier,
    "Reverse": bv_reverse_simplifier,
    "Extract": extract_simplifier,
    "Concat": concat_simplifier,
    "If": if_simplifier,
    "__lshift__": lshift_simplifier,
    "__rshift__": rshift_simplifier,
    "LShR": lshr_simplifier,
    "__eq__": eq_simplifier,
    "__ne__": ne_simplifier,
    "__ge__": ge_simplifier,
    "__or__": bitwise_or_simplifier,
    "__and__": bitwise_and_simplifier,
    "__xor__": bitwise_xor_simplifier,
    "__add__": bitwise_add_simplifier,
    "__sub__": bitwise_sub_simplifier,
    "__mul__": bitwise_mul_simplifier,
    "ZeroExt": zeroext_simplifier,
    "SignExt": signext_simplifier,
    "fpToIEEEBV": fptobv_simplifier,
    "fpToFP": fptofp_simplifier,
    "StrReverse": str_reverse_simplifier,
    "__invert__": invert_simplifier,
}

# the actual function


def simplify(op, args) -> tuple[Base | None, bool]:
    """Simplifies the given operation with the given arguments. Returns the
    simplified result if possible, along with a boolean representing whether
    annotations were handled by the simplifier.
    """

    if op not in _all_simplifiers:
        return None, False

    simplified = _all_simplifiers[op](*args)

    if isinstance(simplified, tuple):
        return simplified[0], simplified[1]

    return simplified, False
